import numpy as np
from colormath.color_objects import LabColor, XYZColor, sRGBColor
from colormath.color_conversions import convert_color
from colormath.color_diff import delta_e_cie1976, delta_e_cie1994, delta_e_cie2000
import util.color_constants as color_constants
import sys

d50 = [0.4931, 0.5651, 0.6003, 0.5782, 0.7482, 0.8725, 0.9061, 0.9137, 0.9511, 0.9196, 0.9572, 0.9661, 0.9713, 1.021, 1.0075, 1.0232, 1., 0.9774, 0.9892, 0.935, 0.9769, 0.9927, 0.9904, 0.9572, 0.9886, 0.9567, 0.9819, 1.03, 0.9913, 0.8738, 0.916]
d50 = np.array(d50)

def LightingMatrix_2Degree(illuminant):
    # color matching functions
    x2 = np.array([0.01, 0.04, 0.13, 0.28, 0.35, 0.34, 0.29, 0.20, 0.10, 0.03, 0.00, 0.01, 0.06, 0.17, 0.29, 0.43, 0.59, 0.76, 0.92, 1.03, 1.06, 1.00, 0.85, 0.64, 0.45, 0.28, 0.16, 0.09, 0.05, 0.02, 0.01])
    y2 = np.array([0.00, 0.00, 0.00, 0.01, 0.02, 0.04, 0.06, 0.09, 0.14, 0.21, 0.32, 0.50, 0.71, 0.86, 0.95, 0.99, 1.00, 0.95, 0.87, 0.76, 0.63, 0.50, 0.38, 0.27, 0.18, 0.11, 0.06, 0.03, 0.02, 0.01, 0.00])
    z2 = np.array([0.07, 0.21, 0.65, 1.39, 1.75, 1.77, 1.67, 1.29, 0.81, 0.47, 0.27, 0.16, 0.08, 0.04, 0.02, 0.01, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00])

    S = 1. / np.sum(illuminant * y2);
    #print(S.shape, illuminant.shape, x2.shape, (illuminant * x2).shape)
    Lh = np.array([S * illuminant * x2, S * illuminant * y2, S * illuminant * z2])
    return Lh

def LABfromSpectra(spectra, illuminant=d50):
    assert(len(spectra.shape) == 1 or len(spectra.shape) == 2)
    if len(spectra.shape) == 1:
        Lh = LightingMatrix_2Degree(illuminant)
        XYZ = np.transpose(np.dot(Lh, spectra))
        #WP = np.transpose(np.dot(Lh, np.ones([31,1])))
        #print(Lh.shape, WP.shape, XYZ.shape)
        #xyz2lab = makecform('xyz2lab','WhitePoint', WP)
        #LAB = applycform(XYZ, xyz2lab)
        XYZ = XYZColor(XYZ[0], XYZ[1], XYZ[2], observer='2', illuminant='d50')
        LAB = convert_color(XYZ, LabColor)
        return np.array(LAB.get_value_tuple())
    else:
        LABs = np.ndarray(shape=[spectra.shape[0],3])
        for i in range(spectra.shape[0]):
            Lh = LightingMatrix_2Degree(illuminant)
            XYZ = np.transpose(np.dot(Lh, spectra[i].T))
            XYZ = XYZColor(XYZ[0], XYZ[1], XYZ[2], observer='2', illuminant='d50')
            LAB = convert_color(XYZ, LabColor)
            LABs[i] = np.array(LAB.get_value_tuple())
        return LABs
def LABfromSpectra_arrayMode(spectra, illuminant='d50', observer='2'):
    assert (illuminant == 'd50')
    illuminant_arr = d50

    # Get XYZ -- from Tejas' MATLAB implementation
    Lh = np.transpose(LightingMatrix_2Degree(illuminant_arr))
    XYZ = np.dot(spectra, Lh)

    # Get LAB -- modified based on colormath's implementation
    illum = color_constants.ILLUMINANTS[observer][illuminant]
    XYZ_tmp = np.divide(XYZ, illum)
    XYZ_tmp = np.where(XYZ_tmp > color_constants.CIE_E, np.power(XYZ_tmp, (1.0 / 3.0)),
                       (7.787 * XYZ_tmp) + (16.0 / 116.0))
    temp_x, temp_y, temp_z = XYZ_tmp[:, 0], XYZ_tmp[:, 1], XYZ_tmp[:, 2]

    lab_l = (116.0 * temp_y) - 16.0
    lab_a = 500.0 * (temp_x - temp_y)
    lab_b = 200.0 * (temp_y - temp_z)

    LABs = np.hstack([np.reshape(lab_l, [-1, 1]), np.reshape(lab_a, [-1, 1]), np.reshape(lab_b, [-1, 1])])

    return LABs
def get_lab_from_spectral(spectral):
    return LABfromSpectra(spectral)



def LAB2sRGB(lab):
    assert(lab.shape == (3,))
    lab_obj = LabColor(lab[0], lab[1], lab[2], observer='2', illuminant='d50')
    sRGB = convert_color(lab_obj, sRGBColor)
    sRGB = np.array([sRGB.rgb_r, sRGB.rgb_g, sRGB.rgb_b])
    return sRGB

def LAB2sRGBImg(lab, black_edge=False):
    sRGB = LAB2sRGB(lab)
    img = sRGB
    img = np.reshape(img, [1,1,3])
    side_len = 16
    img = np.tile(img, [side_len,side_len,1])
    if black_edge:
        img[0,:,:] = 0
        img[side_len-1, :, :] = 0
        img[:, 0, :] = 0
        img[:, side_len - 1, :] = 0
    return img


def my_delta_e_cie2000_arrayVersion(lab_color_vector, lab_color_matrix, Kl=1, Kc=1, Kh=1):
    """
    Calculates the Delta E (CIE2000) of two colors.
    """
    L, a, b = lab_color_vector[:,0], lab_color_vector[:,1], lab_color_vector[:,2]

    avg_Lp = (L + lab_color_matrix[:, 0]) / 2.0

    C1 = np.sqrt(np.sum(np.power(lab_color_vector[:, 1:], 2), axis=1))
    C2 = np.sqrt(np.sum(np.power(lab_color_matrix[:, 1:], 2), axis=1))

    avg_C1_C2 = (C1 + C2) / 2.0

    G = 0.5 * (
        1
        - np.sqrt(
            np.power(avg_C1_C2, 7.0)
            / (np.power(avg_C1_C2, 7.0) + np.power(25.0, 7.0))
        )
    )

    a1p = (1.0 + G) * a
    a2p = (1.0 + G) * lab_color_matrix[:, 1]

    C1p = np.sqrt(np.power(a1p, 2) + np.power(b, 2))
    C2p = np.sqrt(np.power(a2p, 2) + np.power(lab_color_matrix[:, 2], 2))

    avg_C1p_C2p = (C1p + C2p) / 2.0

    h1p = np.degrees(np.arctan2(b, a1p))
    h1p += (h1p < 0) * 360

    h2p = np.degrees(np.arctan2(lab_color_matrix[:, 2], a2p))
    h2p += (h2p < 0) * 360

    avg_Hp = (((np.fabs(h1p - h2p) > 180) * 360) + h1p + h2p) / 2.0

    T = (
        1
        - 0.17 * np.cos(np.radians(avg_Hp - 30))
        + 0.24 * np.cos(np.radians(2 * avg_Hp))
        + 0.32 * np.cos(np.radians(3 * avg_Hp + 6))
        - 0.2 * np.cos(np.radians(4 * avg_Hp - 63))
    )

    diff_h2p_h1p = h2p - h1p
    delta_hp = diff_h2p_h1p + (np.fabs(diff_h2p_h1p) > 180) * 360
    delta_hp -= (h2p > h1p) * 720

    delta_Lp = lab_color_matrix[:, 0] - L
    delta_Cp = C2p - C1p
    delta_Hp = 2 * np.sqrt(C2p * C1p) * np.sin(np.radians(delta_hp) / 2.0)

    S_L = 1 + (
        (0.015 * np.power(avg_Lp - 50, 2))
        / np.sqrt(20 + np.power(avg_Lp - 50, 2.0))
    )
    S_C = 1 + 0.045 * avg_C1p_C2p
    S_H = 1 + 0.015 * avg_C1p_C2p * T

    delta_ro = 30 * np.exp(-(np.power(((avg_Hp - 275) / 25), 2.0)))
    R_C = np.sqrt(
        (np.power(avg_C1p_C2p, 7.0))
        / (np.power(avg_C1p_C2p, 7.0) + np.power(25.0, 7.0))
    )
    R_T = -2 * R_C * np.sin(2 * np.radians(delta_ro))

    Elabs = np.sqrt(
        np.power(delta_Lp / (S_L * Kl), 2)
        + np.power(delta_Cp / (S_C * Kc), 2)
        + np.power(delta_Hp / (S_H * Kh), 2)
        + R_T * (delta_Cp / (S_C * Kc)) * (delta_Hp / (S_H * Kh))
    )
    return np.reshape(Elabs, [Elabs.shape[0], 1])


def color_diff(lab1, lab2, type='cie2000', Kl=1, Kc=1, Kh=1):
    assert(lab1.shape == lab2.shape)
    assert(lab1.shape[-1] == 3)
    assert(len(lab1.shape) == 1 or len(lab1.shape) == 2)
    if len(lab1.shape) == 1:
        color1 = LabColor(lab_l=lab1[0], lab_a=lab1[1], lab_b=lab1[2])
        color2 = LabColor(lab_l=lab2[0], lab_a=lab2[1], lab_b=lab2[2])
        if type == 'cie2000':
            color_diff = delta_e_cie2000(color1, color2, Kl=Kl, Kc=Kc, Kh=Kh)
        elif type == 'cie1994':
            color_diff = delta_e_cie1994(color1, color2, K_L=Kl, K_C=Kc, K_H=Kh)
        elif type == 'cie1976':
            color_diff = delta_e_cie1976(color1, color2)
        else:
            sys.exit('unknown color type!')
        return color_diff
    else:
        if type == 'cie2000':
            return my_delta_e_cie2000_arrayVersion(lab1, lab2, Kl=Kl, Kc=Kc, Kh=Kh)
        else:
            color_diffs = np.ndarray(shape=[lab1.shape[0],1])
            for i in range(lab1.shape[0]):
                color1 = LabColor(lab_l=lab1[i, 0], lab_a=lab1[i, 1], lab_b=lab1[i, 2])
                color2 = LabColor(lab_l=lab2[i, 0], lab_a=lab2[i, 1], lab_b=lab2[i, 2])
                if type == 'cie1994':
                    color_diffs[i] = delta_e_cie1994(color1, color2, K_L=Kl, K_C=Kc, K_H=Kh)
                elif type == 'cie1976':
                    color_diffs[i] = delta_e_cie1976(color1, color2)
                else:
                    sys.exit('unknown color type!')
            return color_diffs

